//
//  OTPSOCK
//
//	Copyright (C) 1996-1998 NEC Corporation
//	All rights reserved.
//  Authos(s): Mizukoshi, Yasuhiro <mizukosi@ccs.mt.nec.co.jp>
//
//

static char CopyRightID[] = "Copyright(C) 1996-1998 NEC Corporation. All rights reserved.";

//	    $Archive: /sotools/otpsock/Otpsock.c $
//			$Revision: 18 $
//			$Date: 98/10/22 10:18p $
static char VssID[] = "$Header: /sotools/otpsock/Otpsock.c 18    98/10/22 10:18p Mizukosi $";

//
//	Functions:
//		connect
//		recv
//		send
//		DllMain
//		LoadWinsock
//		FreeWinsock
//		PassphraseDlgProc
//		SendPassMode
//		SetFuncName
//		chkResp
//		otpmain
//


#include <windows.h>
#include <winsock.h>

#include <stdio.h>

#define	GLOBAL
#include "otpsock.h"
#include "resource.h"


///////////////////////////////////////////
//
//	OTP Response Keyword
//
struct KeywordPattern keys[] = {
	{ "s/key",     MD4,  TRUE  },
	{ "otp-md4",   MD4,  FALSE },
	{ "otp-md5",   MD5,  FALSE },
	{ "otp-sha1",  SHA1, FALSE },
	{ "S/Key MD4", MD4,  TRUE  },
	{ "S/Key MD5", MD5,  TRUE  },
	{ "Skey Challenge: s/key", MD4, TRUE },
	{ NULL, 0, FALSE },
};
struct KeywordPattern ftpkeys[] = {
	{ "331 s/key",    MD4,  TRUE  },
	{ "331 otp-md4",  MD4,  FALSE },
	{ "331 otp-md5",  MD5,  FALSE },
	{ "331 otp-sha1", SHA1, FALSE },
	{ "331 OTP response otp-md4",  MD4,  FALSE },
	{ "331 OTP response otp-md5",  MD5,  FALSE },
	{ "331 OTP response otp-sha1", SHA1, FALSE },
	{ "331 Response to otp-md4",   MD4,  FALSE },
	{ "331 Response to otp-md5",   MD5,  FALSE },
	{ "331 Response to otp-sha1",  SHA1, FALSE },
	{ NULL, 0, FALSE },
};
struct KeywordPattern *optkeys = NULL;


////////////////////////////////////////////////
//	connect
//
int PASCAL FAR EXPORT
connect (SOCKET s, const struct sockaddr FAR *name, int namelen)
{
	int		ret;

x_OutputDebugString("connect ");

	// Detect sending password mode
	if ( gOTPData.mode == 0 )
		SendPassMode(name, namelen);

	ret = gft.connect(s, name, namelen);

	return ret;
}


/////////////////////////////////////////////////
//	recv
//
int PASCAL FAR EXPORT
recv (SOCKET s, char FAR * buf, int len, int flags)
{
	int		ret;
	struct LocalData *ld;

	ld = (struct LocalData *)TlsGetValue(TlsIndex);

x_OutputDebugString("recv ");

	if (!ld->bInCall){
		ld->bInCall = TRUE;

		ret = gft.recv(s, buf, len, flags);

		// scan data
		if (ret != SOCKET_ERROR)
			otpmain(s, buf, ret, flags);

		ld->bInCall = FALSE;

	} else {
		// recursive call
		ret = gft.recv(s, buf, len, flags);
	}

	return ret;
}


/////////////////////////////////////////////
//	send
//
int PASCAL FAR EXPORT
send (SOCKET s, const char FAR * buf, int len,
    int flags)
{
	int		ret;
	struct LocalData *ld;

	ld = (struct LocalData *)TlsGetValue(TlsIndex);

x_OutputDebugString("send ");

	// If ftp-mode is set, passphrase is sent before sending data.
	if (gOTPData.bSendPass && !ld->bInCall){
		char *p;
		int  sent, total;
		char sbuf[ONETIME_KEY_MAX_LEN + 8];

		wsprintf(sbuf, "PASS %s\r\n", gOTPData.pass);

		p     = sbuf;
		total = strlen(sbuf);
		do {
			ld->bInCall = TRUE;
			sent = gft.send(s, p, total, flags);
			ld->bInCall = FALSE;

			total -= sent;
			p     += sent;
		} while (total) ;

		gOTPData.bSendPass = FALSE;

		return len;
	}

	ret = gft.send(s, buf, len, flags);

	return ret;
}


////////////////////////////////////////////////////////
//	WSAStartup
//
int PASCAL FAR EXPORT
WSAStartup(WORD wVersionRequired, LPWSADATA lpWSAData)
{
	int ret;
	
x_OutputDebugString("WSAStartup ");

	if (gft.hWinsock == NULL){
		if (LoadWinsock() == FALSE){
			return WSASYSNOTREADY;
		}
	}

	ret = gft.WSAStartup(wVersionRequired, lpWSAData);
	
	return ret;
}


///////////////////////////////////////////////////
//	OTPSOCK Entry
//
BOOL APIENTRY DllMain(HANDLE hModule, DWORD ul_reason_for_call, LPVOID lpReserved)
{
	struct LocalData		*ld;
	int i;

	switch(ul_reason_for_call){
	case DLL_PROCESS_ATTACH:
		ghModule = hModule;
		TlsIndex = TlsAlloc();

		gft.hWinsock = NULL;

		InitPreferences();

		ld = (struct LocalData *)LocalAlloc(LPTR, sizeof(struct LocalData));
		TlsSetValue(TlsIndex, (LPVOID)ld);
		ld->hWinsock    = gft.hWinsock;
		ld->bInCall     = FALSE;

		gOTPData.recvbuf_pos = 0;
		gOTPData.mode        = 0;
		gOTPData.bSendPass   = FALSE;
		break;

	case DLL_THREAD_ATTACH:
		ld = (struct LocalData *)LocalAlloc(LPTR, sizeof(struct LocalData));
		TlsSetValue(TlsIndex, (LPVOID)ld);
		ld->hWinsock    = LoadLibrary(ORIGINAL_WSOCK32);
		ld->bInCall     = FALSE;
		break;

	case DLL_THREAD_DETACH:
		ld = (struct LocalData *)TlsGetValue(TlsIndex);
		FreeLibrary(ld->hWinsock);
		LocalFree(ld);
		break;

	case DLL_PROCESS_DETACH:
		ld = (struct LocalData *)TlsGetValue(TlsIndex);
		LocalFree(ld);

		if (optkeys){
			for (i = 0; optkeys[i].key != NULL; i++)
				free(optkeys[i].key);
			free(optkeys);
		}

		FreeWinsock();

		TlsFree(TlsIndex);
		break;
	}

	return TRUE;
}


////////////////////////////////////////////////////////
//
//	LoadWinsock
//
int LoadWinsock()
{
	gft.hWinsock = LoadLibrary(ORIGINAL_WSOCK32);

	if (gft.hWinsock == NULL){
		char szWSPath[MAX_PATH];

		if (!GetSystemDirectory(szWSPath, MAX_PATH))
			return FALSE;
		
		strcat(szWSPath, "\\wsock32.dll");

		gft.hWinsock = LoadLibrary(szWSPath);
	}

	if (gft.hWinsock == NULL)
		return FALSE;

	gft.accept                = (void *)GetProcAddress(gft.hWinsock, "accept");
	gft.bind                  = (void *)GetProcAddress(gft.hWinsock, "bind");
	gft.closesocket           = (void *)GetProcAddress(gft.hWinsock, "closesocket");
	gft.connect               = (void *)GetProcAddress(gft.hWinsock, "connect");
	gft.getpeername           = (void *)GetProcAddress(gft.hWinsock, "getpeername");
	gft.getsockname           = (void *)GetProcAddress(gft.hWinsock, "getsockname");
	gft.getsockopt            = (void *)GetProcAddress(gft.hWinsock, "getsockopt");
	gft.htonl                 = (void *)GetProcAddress(gft.hWinsock, "htonl");
	gft.htons                 = (void *)GetProcAddress(gft.hWinsock, "htons");
	gft.inet_addr             = (void *)GetProcAddress(gft.hWinsock, "inet_addr");
	gft.inet_ntoa             = (void *)GetProcAddress(gft.hWinsock, "inet_ntoa");
	gft.ioctlsocket           = (void *)GetProcAddress(gft.hWinsock, "ioctlsocket");
	gft.listen                = (void *)GetProcAddress(gft.hWinsock, "listen");
	gft.ntohl                 = (void *)GetProcAddress(gft.hWinsock, "ntohl");
	gft.ntohs                 = (void *)GetProcAddress(gft.hWinsock, "ntohs");
	gft.recv                  = (void *)GetProcAddress(gft.hWinsock, "recv");
	gft.recvfrom              = (void *)GetProcAddress(gft.hWinsock, "recvfrom");
	gft.select                = (void *)GetProcAddress(gft.hWinsock, "select");
	gft.send                  = (void *)GetProcAddress(gft.hWinsock, "send");
	gft.sendto                = (void *)GetProcAddress(gft.hWinsock, "sendto");
	gft.setsockopt            = (void *)GetProcAddress(gft.hWinsock, "setsockopt");
	gft.socket                = (void *)GetProcAddress(gft.hWinsock, "socket");
	gft.shutdown              = (void *)GetProcAddress(gft.hWinsock, "shutdown");

	gft.gethostbyaddr         = (void *)GetProcAddress(gft.hWinsock, "gethostbyaddr");
	gft.gethostbyname         = (void *)GetProcAddress(gft.hWinsock, "gethostbyname");
	gft.getprotobyname        = (void *)GetProcAddress(gft.hWinsock, "getprotobyname");
	gft.getprotobynumber      = (void *)GetProcAddress(gft.hWinsock, "getprotobynumber");
	gft.getservbyname         = (void *)GetProcAddress(gft.hWinsock, "getservbyname");
	gft.getservbyport         = (void *)GetProcAddress(gft.hWinsock, "getservbyport");
	gft.gethostname           = (void *)GetProcAddress(gft.hWinsock, "gethostname");

	gft.WSAAsyncSelect        = (void *)GetProcAddress(gft.hWinsock, "WSAAsyncSelect");
	gft.WSAAsyncGetHostByAddr = (void *)GetProcAddress(gft.hWinsock, "WSAAsyncGetHostByAddr");
	gft.WSAAsyncGetHostByName = (void *)GetProcAddress(gft.hWinsock, "WSAAsyncGetHostByName");
	gft.WSAAsyncGetProtoByNumber = (void *)GetProcAddress(gft.hWinsock, "WSAAsyncGetProtoByNumber");
	gft.WSAAsyncGetProtoByName = (void *)GetProcAddress(gft.hWinsock, "WSAAsyncGetProtoByName");
	gft.WSAAsyncGetServByPort = (void *)GetProcAddress(gft.hWinsock, "WSAAsyncGetServByPort");
	gft.WSAAsyncGetServByName = (void *)GetProcAddress(gft.hWinsock, "WSAAsyncGetServByName");
	gft.WSACancelAsyncRequest = (void *)GetProcAddress(gft.hWinsock, "WSACancelAsyncRequest");
	gft.WSASetBlockingHook    = (void *)GetProcAddress(gft.hWinsock, "WSASetBlockingHook");
	gft.WSAUnhookBlockingHook = (void *)GetProcAddress(gft.hWinsock, "WSAUnhookBlockingHook");
	gft.WSAGetLastError       = (void *)GetProcAddress(gft.hWinsock, "WSAGetLastError");
	gft.WSASetLastError       = (void *)GetProcAddress(gft.hWinsock, "WSASetLastError");
	gft.WSACancelBlockingCall = (void *)GetProcAddress(gft.hWinsock, "WSACancelBlockingCall");
	gft.WSAIsBlocking         = (void *)GetProcAddress(gft.hWinsock, "WSAIsBlocking");
	gft.WSAStartup            = (void *)GetProcAddress(gft.hWinsock, "WSAStartup");
	gft.WSACleanup            = (void *)GetProcAddress(gft.hWinsock, "WSACleanup");

	gft.__WSAFDIsSet          = (void *)GetProcAddress(gft.hWinsock, "__WSAFDIsSet");

	gft.EnumProtocolsA		  = (void *)GetProcAddress(gft.hWinsock, "EnumProtocolsA");
	gft.EnumProtocolsW		  = (void *)GetProcAddress(gft.hWinsock, "EnumProtocolsW");
	gft.GetAddressByNameA	  = (void *)GetProcAddress(gft.hWinsock, "GetAddressByNameA");
	gft.GetAddressByNameW	  = (void *)GetProcAddress(gft.hWinsock, "GetAddressByNameW");
	gft.GetNameByTypeA		  = (void *)GetProcAddress(gft.hWinsock, "GetNameByTypeA");
	gft.GetNameByTypeW		  = (void *)GetProcAddress(gft.hWinsock, "GetNameByTypeW");
	gft.GetServiceA			  = (void *)GetProcAddress(gft.hWinsock, "GetServiceA");
	gft.GetServiceW			  = (void *)GetProcAddress(gft.hWinsock, "GetServiceW");
	gft.GetTypeByNameA		  = (void *)GetProcAddress(gft.hWinsock, "GetTypeByNameA");
	gft.GetTypeByNameW		  = (void *)GetProcAddress(gft.hWinsock, "GetTypeByNameW");
	gft.SetServiceA			  = (void *)GetProcAddress(gft.hWinsock, "SetServiceA");
	gft.SetServiceW			  = (void *)GetProcAddress(gft.hWinsock, "SetServiceW");
	gft.TransmitFile		  = (void *)GetProcAddress(gft.hWinsock, "TransmitFile");
	gft.WSARecvEx			  = (void *)GetProcAddress(gft.hWinsock, "WSARecvEx");

	gft.WsControl             = (void *)GetProcAddress(gft.hWinsock, "WsControl");
	gft.closesockinfo         = (void *)GetProcAddress(gft.hWinsock, "closesockinfo");
	gft.Arecv                 = (void *)GetProcAddress(gft.hWinsock, "Arecv");
	gft.Asend                 = (void *)GetProcAddress(gft.hWinsock, "Asend");
	gft.WSHEnumProtocols      = (void *)GetProcAddress(gft.hWinsock, "WSHEnumProtocols");
	gft.inet_network          = (void *)GetProcAddress(gft.hWinsock, "inet_network");
	gft.getnetbyname          = (void *)GetProcAddress(gft.hWinsock, "getnetbyname");
	gft.rcmd                  = (void *)GetProcAddress(gft.hWinsock, "rcmd");
	gft.rexec                 = (void *)GetProcAddress(gft.hWinsock, "rexec");
	gft.rresvport             = (void *)GetProcAddress(gft.hWinsock, "rresvport");
	gft.sethostname           = (void *)GetProcAddress(gft.hWinsock, "sethostname");
	gft.dn_expand             = (void *)GetProcAddress(gft.hWinsock, "dn_expand");
	gft.s_perror              = (void *)GetProcAddress(gft.hWinsock, "s_perror");
	gft.NPLoadNameSpaces      = (void *)GetProcAddress(gft.hWinsock, "NPLoadNameSpaces");

x_OutputDebugString("END: LoadWinsock\n");

	return TRUE;
}


//////////////////////////////////////////////////////////////
//
//	FreeWinsock
//
int FreeWinsock()
{
	FreeLibrary(gft.hWinsock);

x_OutputDebugString("FreeWinsock\n");

	return TRUE;
}


//////////////////////////////////////////////////////////////
//
//	Passphrase Input Dialog Procedure
//
BOOL CALLBACK PassphraseDlgProc(HWND hWnd, UINT uMessage,
				WPARAM wParam, LPARAM lParam)
{
    switch(uMessage) {
    case WM_INITDIALOG:
	{
		char Title[SEED_MAX_LEN + 10];
		POINT pt, spt;
		RECT  r;

		wsprintf(Title, "%d %s", gOTPData.seq, gOTPData.seed);
		SetWindowText(hWnd, Title);		

		SetWindowText(GetDlgItem(hWnd, IDC_FUNC), gOTPData.FuncName);
		CheckDlgButton(hWnd, IDC_TELNET, (gOTPData.mode - FTP_MODE));

		GetCursorPos(&pt);
		GetWindowRect(hWnd, &r);
		spt.x = GetSystemMetrics(SM_CXSCREEN) - pt.x - (r.right - r.left);
		spt.y = GetSystemMetrics(SM_CYSCREEN) - pt.y - (r.bottom - r.top);
		if (spt.x < 0) pt.x += spt.x;
		if (spt.y < 0) pt.y += spt.y;

		MoveWindow(hWnd, pt.x, pt.y, r.right - r.left, r.bottom - r.top, TRUE);
		ShowWindow(hWnd, SW_SHOWNORMAL);
		SetForegroundWindow(hWnd);
	}
		break;

	case WM_COMMAND:
		switch(wParam){
		case IDOK:
		{
			char passphrase[PASS_PHRASE_MAX_LEN];
			char key[8];
			int  i;

			GetDlgItemText(hWnd, IDC_PASS, passphrase, PASS_PHRASE_MAX_LEN);
			gOTPData.mode = IsDlgButtonChecked(hWnd, IDC_TELNET) + FTP_MODE;

			if (gOTPData.KeyCrunchFunc(key, gOTPData.seed, passphrase) == -1){
				memset(passphrase, 0, PASS_PHRASE_MAX_LEN);
				EndDialog(hWnd, FALSE);
				break;
			}
			for (i = 0; i < gOTPData.seq; i++) gOTPData.IterateFunc(key);
			OTP_btoe(gOTPData.pass, key);

			memset(passphrase, 0, PASS_PHRASE_MAX_LEN);
			EndDialog(hWnd, TRUE);
		}
			break;

		case IDCANCEL:
			EndDialog(hWnd, FALSE);
			break;

		default:
			return FALSE;
		}
		break;

	default:
		return FALSE;
	}
	return TRUE;
}																	   	


////////////////////////////////////////////////
//
//	Check Port Number
//
int SendPassMode(const struct sockaddr FAR *name, int namelen)
{
	short	port;

	if (namelen == sizeof(struct sockaddr_in)){
		port = ntohs(((struct sockaddr_in *)name)->sin_port);

		if (port == IPPORT_FTP ||
			port == 110 /* POP3 */){
			gOTPData.mode = FTP_MODE;
			return TRUE;
		}
	}

	gOTPData.mode = TELNET_MODE;

	return TRUE;
}


///////////////////////////////////////////
//
//	Set Function Name & Pointers
//
int SetFuncName()
{
	switch(gOTPData.hash_func){
	case MD4:
		gOTPData.KeyCrunchFunc = (void *)_otp_MD4_keycrunch;
		gOTPData.IterateFunc   = (void *)_otp_MD4_iterate;
		strcpy(gOTPData.FuncName, FUNCNAME_MD4);
		break;
	case MD5:
		gOTPData.KeyCrunchFunc = (void *)_otp_MD5_keycrunch;
		gOTPData.IterateFunc   = (void *)_otp_MD5_iterate;
		strcpy(gOTPData.FuncName, FUNCNAME_MD5);
		break;
	case SHA1:
		gOTPData.KeyCrunchFunc = (void *)_otp_SHA1_keycrunch;
		gOTPData.IterateFunc   = (void *)_otp_SHA1_iterate;
		strcpy(gOTPData.FuncName, FUNCNAME_SHA1);
		break;

	default:
		return FALSE;
	}

	return TRUE;
}


////////////////////////////////////////
//
//	Keyword Search
//
int KeywordSearch(struct KeywordPattern *kp)
{
	int		i;

	if (kp == NULL)
		return FALSE;

	for (i = 0; kp[i].key != NULL; i++){

		if ( !strncmp(gOTPData.recvbuf, kp[i].key, strlen(kp[i].key)) ){
			gOTPData.hash_func = kp[i].func;

			// for S/Key MD5
			if ( kp[i].IsSKey && gbSKeyMD5 )
				gOTPData.hash_func = MD5;

			SetFuncName();

			if ( sscanf(gOTPData.recvbuf + strlen(kp[i].key), SEQSEED_POS, &gOTPData.seq, gOTPData.seed) != 2 ){
				return FALSE;
			}

			return TRUE;

		}
	}

	return FALSE;
}


////////////////////////////////////////
//
//	Check Response
//
int chkResp()
{
	struct KeywordPattern *kp;

	if ( !strncmp(gOTPData.recvbuf, FTP_RESPONSE, strlen(FTP_RESPONSE)) ){
		kp = ftpkeys;
	} else {
		kp = keys;
	}

	if (KeywordSearch(kp) || KeywordSearch(optkeys))
		return TRUE;
	else
		return FALSE;
}


/////////////////////////////////////////////////////////
//
//	OTP Main
//		SOCKET		s,		socket
//		char FAR	*buf,	receive data buffer
//		int			len		receive data length
//		int			flags	flag of send()
//
int otpmain(SOCKET s, char FAR * buf, int len, int flags)
{
	int		i;
	int		res;

	for (i = 0; i < len; i++){

		// Over flow
		if (gOTPData.recvbuf_pos >= MAX_OTPBUF_SIZE){
			gOTPData.recvbuf[0]  = 0;
			gOTPData.recvbuf_pos = 0;
		}

		gOTPData.recvbuf[gOTPData.recvbuf_pos] = buf[i];

		if (gOTPData.recvbuf[gOTPData.recvbuf_pos] == '\n'){
			// Set NULL for end-of-string
			gOTPData.recvbuf[gOTPData.recvbuf_pos] = 0;
			gOTPData.recvbuf_pos = 0;

		} else if (iReqEndLen &&
				   iReqEndLen <= (gOTPData.recvbuf_pos + 1) &&
				   !strncmp((gOTPData.recvbuf+gOTPData.recvbuf_pos+1-iReqEndLen), szReqEnd, iReqEndLen)){
			gOTPData.recvbuf[gOTPData.recvbuf_pos+1] = 0;
			gOTPData.recvbuf_pos++;

		} else {
			gOTPData.recvbuf_pos++;
			continue;
		}

		// Chech whether recv data includes challenges.
		if ( chkResp() ){
			HWND	hWndDotkey;

			// Find OTP calculater
			hWndDotkey = FindWindow(NULL, gOTPCalcWndName);

			if (hWndDotkey){

				res = FALSE;

				wsprintf(gtbuf, OTPCALC_SERVER_TMPL,
						gOTPData.FuncName,
						gOTPData.seq,
						gOTPData.seed);

				gCDS.dwData = 0;
				gCDS.cbData = MAX_OTPBUF_SIZE + 1;
				gCDS.lpData = gtbuf;

				// Copy challenges to OTP calculater
				SendMessage(hWndDotkey, WM_COPYDATA,
							(WPARAM)NULL, (LPARAM)&gCDS);

				Sleep(500);

				// Get result from OTP calculater
				if (OpenClipboard(NULL) == TRUE){
					HANDLE hClipMem = GetClipboardData(CF_TEXT);

					if (hClipMem != NULL){
						char * pClipMem = (char *)GlobalLock(hClipMem);

						if (pClipMem[0]){
							strcpy(gOTPData.pass, pClipMem);

							res = TRUE;
						}

						GlobalUnlock(hClipMem);
					}
					CloseClipboard();
				}

			} else {
				CheckSeqNum();

				res = DialogBox(ghModule, MAKEINTRESOURCE(IDD_PASSPHRASE),
								NULL, PassphraseDlgProc);

			}

			if (res){
				char *p;
				int  sent, total;
				char sbuf[ONETIME_KEY_MAX_LEN + 8];

				if (gOTPData.mode == FTP_MODE){
					gOTPData.bSendPass = TRUE;

				} else {
					wsprintf(sbuf, "%s\r\n", gOTPData.pass);

					p     = sbuf;
					total = strlen(sbuf);
					do {
						sent = gft.send(s, p, total, flags);

						total -= sent;
						p     += sent;
					} while (total) ;
				}
			}
		}
	}

	return TRUE;
}


/////////////////////////////////////////////////////////
//
//	Check Sequence Number
//
int CheckSeqNum()
{
	char szbuf[128];

	switch (gOTPData.seq){
	case 20: case 10: case  5: case  4:
	case  3: case  2: case  1: case  0:
		wsprintf(szbuf, "***** Warning!!  %d times to go. *****", gOTPData.seq);
		MessageBox(NULL, szbuf, "otpsock", MB_OK | MB_SYSTEMMODAL);
		break;
	default:
		break;
	}

	return TRUE;
}


////////////////////////////////////////////////////////////////
//
//	Initialize Preferences
//
int	InitPreferences()
{
	HKEY  hKey;
	DWORD dwType;
	DWORD dwLength;
	DWORD dwData;
	int   i, ret, cbValue;
	char  szValue[MAX_KEYWORD_LEN];

	/////////////////////////////////
	// Set Default Value
	strcpy(gOTPCalcWndName, DEFAULT_OTPCALC);
	gbSKeyMD5   = FALSE;
	szReqEnd[0] = 0;
	iReqEndLen  = strlen(szReqEnd);


	/////////////////////////////////
	// Load Registory
	if (RegOpenKeyEx(HKEY_CURRENT_USER, OTPSOCK_SUBKEY, 0, KEY_QUERY_VALUE, &hKey) != ERROR_SUCCESS){
		return FALSE;
	}

	// OTPCalcWndName
	dwType   = REG_SZ;
	dwLength = sizeof(gOTPCalcWndName);
	RegQueryValueEx(hKey, OTPCALC_VALNAME, NULL, &dwType, gOTPCalcWndName, &dwLength);

	// bSKeyMD5
	dwType   = REG_DWORD;
	dwLength = sizeof(dwData);
	if (RegQueryValueEx(hKey, SKEY_MD5_VALNAME, NULL, &dwType, (LPBYTE)&dwData, &dwLength) == ERROR_SUCCESS)
		gbSKeyMD5 = (BOOL)dwData;

	// ChallengeEndChar
	dwType	 = REG_SZ;
	dwLength = sizeof(szReqEnd);
	RegQueryValueEx(hKey, REQ_END_VALNAME, NULL, &dwType, szReqEnd, &dwLength);
	iReqEndLen  = strlen(szReqEnd);

	RegCloseKey(hKey);


	///////////////////////////////////////////
	// Load optional challenge keywords
	if (RegOpenKeyEx(HKEY_CURRENT_USER, OTPSOCK_KEYWORDS_SUBKEY, 0, KEY_QUERY_VALUE, &hKey) != ERROR_SUCCESS){
		return FALSE;
	}

	// Get the number of optional challenge keywords
	i = 0;
	while (1){
		cbValue  = sizeof(szValue);
		ret = RegEnumValue(hKey, i, szValue, &cbValue, NULL, NULL, NULL, NULL);
		if (ret != ERROR_SUCCESS)
			break;
		i++;
	}
	if (i == 0){
		goto error;
	} else {
		optkeys = (struct KeywordPattern *)malloc(sizeof(struct KeywordPattern) * (i + 1));
		if (optkeys == NULL)
			goto error;
	}

	// Set value to optkeys variable
	i = 0;
	while (1){
		cbValue  = sizeof(szValue);
		dwType   = REG_DWORD;
		dwLength = sizeof(dwData);

		ret = RegEnumValue(hKey, i, szValue, &cbValue, NULL, &dwType, (LPBYTE)&dwData, &dwLength);

		if (ret != ERROR_SUCCESS)
			break;

		optkeys[i].key = (char *)malloc(cbValue+1);
		if (optkeys[i].key == NULL)
			break;
		strcpy(optkeys[i].key, szValue);

		if (dwData <= MD2) dwData = MD5;
		if (dwData > SHA1) dwData = MD5;
		optkeys[i].func = dwData;

		optkeys[i].IsSKey = FALSE;

		i++;
	}

	// Set the termination to optkeys variable
	optkeys[i].key = NULL;

error:
	RegCloseKey(hKey);
	return TRUE;
}
